import os
import typing

import numpy as np
import numpy.typing as npt
from rasterio.crs import CRS as _CRS

FilePath: typing.TypeAlias = str | bytes | os.PathLike[str]

RawCRS: typing.TypeAlias = _CRS | str | int | dict[str, typing.Any]

DataDtypes: typing.TypeAlias = (
    np.bool_
    | np.int8
    | np.int16
    | np.int32
    | np.int64
    | np.uint8
    | np.uint16
    | np.uint32
    | np.uint64
    | np.intp
    | np.uintp
    | np.float16
    | np.float32
    | np.float64
)
SupportedDtypes: typing.TypeAlias = int | bool | float | DataDtypes
RasterData: typing.TypeAlias = npt.NDArray[DataDtypes]
RasterMask: typing.TypeAlias = npt.NDArray[np.bool_]
NumpyUFuncMethod: typing.TypeAlias = typing.Literal[
    "__call__",
    "reduce",
    "reduceat",
    "accumulate",
    "outer",
    "at",
]



from typing import Literal

import affine
import numpy as np
from rasterio import merge as rio_merge

from rasterra._array import RasterArray


def validate_property(
    rasters: list[RasterArray],
    property_name: str,
    *,
    is_nan: bool = False,
) -> None:
    ref_property = getattr(rasters[0], property_name)

    def _property_matches(r: RasterArray) -> bool:
        return bool(
            (is_nan and np.isnan(getattr(r, property_name)))
            or getattr(r, property_name) == ref_property
        )  # fmt: skip

    if not all(_property_matches(raster) for raster in rasters):
        msg = f"All rasters must have the same {property_name}."
        raise ValueError(msg)


def merge(
    rasters: list[RasterArray],
    method: Literal["first", "last", "min", "max", "sum", "count"] = "first",
) -> RasterArray:
    """Merge a list of rasters into a single raster."""
    validate_property(rasters, "crs")
    crs = rasters[0].crs
    validate_property(rasters, "dtype")
    dtype = rasters[0].dtype
    no_data_isnan = bool(
        np.issubdtype(dtype, np.floating) and np.isnan(rasters[0].no_data_value)
    )
    validate_property(rasters, "no_data_value", is_nan=no_data_isnan)
    no_data_value = rasters[0].no_data_value
    validate_property(rasters, "resolution")
    x_res, y_res = rasters[0].resolution
    y_res = -y_res  # rasterio uses negative y resolution

    merge_method = rio_merge.MERGE_METHODS[method]

    left, right, bottom, top = zip(*[raster.bounds for raster in rasters], strict=False)
    dst_w, dst_s, dst_e, dst_n = min(left), min(bottom), max(right), max(top)

    output_width = int(round((dst_e - dst_w) / x_res))
    output_height = int(round((dst_n - dst_s) / y_res))

    dest = no_data_value * np.ones((output_height, output_width), dtype=dtype)
    for raster in rasters:
        # Find the row and column offset of the input raster
        row_off = int(round((dst_n - raster.y_max) / y_res))
        col_off = int(round((raster.x_min - dst_w) / x_res))
        window = (
            slice(row_off, row_off + raster.height),
            slice(col_off, col_off + raster.width),
        )
        dest_region = dest[window]
        if np.isnan(no_data_value):
            dest_mask = np.isnan(dest_region)
        elif np.issubdtype(dtype, np.floating):
            dest_mask = np.isclose(dest_region, no_data_value)
        else:
            dest_mask = dest_region == no_data_value

        source_mask = raster.no_data_mask
        merge_method(dest_region, raster.to_numpy(), dest_mask, source_mask)

    dest_transform = affine.Affine(
        a=x_res,
        b=0,
        c=dst_w,
        d=0,
        e=-y_res,
        f=dst_n,
    )
    return RasterArray(
        dest,
        transform=dest_transform,
        crs=crs,
        no_data_value=no_data_value,
    )



import numbers
import typing

import geopandas as gpd
import numpy as np
import numpy.typing as npt
from affine import Affine
from numpy.core.multiarray import flagsobj
from rasterio.crs import CRS
from rasterio.warp import Resampling, reproject
from shapely.geometry import MultiPolygon, Polygon

from rasterra._features import raster_geometry_mask, to_gdf
from rasterra._plotting import Plotter
from rasterra._typing import (
    DataDtypes,
    FilePath,
    NumpyUFuncMethod,
    RasterData,
    RasterMask,
    RawCRS,
    SupportedDtypes,
)

_RESAMPLING_MAP = {data.name: data for data in Resampling}

NO_DATA_UNSET = None

_IDENTITY_TRANSFORM: Affine = Affine.identity()


class RasterArray(np.lib.mixins.NDArrayOperatorsMixin):
    def __init__(
        self,
        data: RasterData,
        transform: Affine = _IDENTITY_TRANSFORM,
        crs: RawCRS | None = None,
        no_data_value: SupportedDtypes | None = NO_DATA_UNSET,
    ):
        """
        Initialize a RasterArray.

        Parameters
        ----------
        data
            2D NumPy array representing raster data.
        transform
            Affine transform to georeference the raster.
        crs
            Coordinate reference system.
        no_data_value
            Value representing no data.

        """
        self._ndarray = data
        self._transform = transform
        if crs is not None and not isinstance(crs, CRS):
            crs = CRS.from_user_input(crs)
        self._crs: CRS | None = crs
        self._no_data_value = no_data_value

    # ----------------------------------------------------------------
    # Array data

    @property
    def flags(self) -> flagsobj:
        """Flags of the raster."""
        return self._ndarray.flags

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the raster."""
        return self._ndarray.shape

    @property
    def strides(self) -> tuple[int, ...]:
        """Strides of the raster."""
        return self._ndarray.strides

    @property
    def ndim(self) -> int:
        """Number of dimensions of the raster."""
        return 2

    @property
    def data(self) -> memoryview:
        """Python buffer object pooint to the start of the raster's data."""
        return self._ndarray.data

    @property
    def size(self) -> int:
        """Number of elements in the raster."""
        return self._ndarray.size

    @property
    def itemsize(self) -> int:
        """Size of each element in the raster."""
        return self._ndarray.itemsize

    @property
    def nbytes(self) -> int:
        """Number of bytes in the raster."""
        return self._ndarray.nbytes

    @property
    def base(self) -> RasterData | None:
        """Base object of the raster."""
        return self._ndarray.base

    @property
    def dtype(self) -> np.dtype[DataDtypes]:
        """Data type of the raster."""
        return self._ndarray.dtype

    @property
    def T(self) -> typing.NoReturn:  # noqa: N802
        """Transpose of the raster."""
        msg = "Transpose of a raster is not defined."
        raise TypeError(msg)

    @property
    def real(self) -> typing.NoReturn:
        """Real part of the raster."""
        msg = "Complex raster data is not supported."
        raise NotImplementedError(msg)

    @property
    def imag(self) -> typing.NoReturn:
        """Imaginary part of the raster."""
        msg = "Complex raster data is not supported."
        raise NotImplementedError(msg)

    @property
    def flat(self) -> np.flatiter:  # type: ignore[type-arg]
        """Flat iterator of the raster."""
        return self._ndarray.flat

    @property
    def ctypes(self) -> typing.NoReturn:
        """ctypes object of the raster."""
        msg = "ctypes object of a raster is not defined."
        raise TypeError(msg)

    def __getitem__(self, item: int | slice | tuple[int, int] | tuple[slice, slice]):  # type: ignore[no-untyped-def]
        def _process_item(_item: int | slice) -> int | slice:
            if isinstance(_item, int):
                return _item
            elif isinstance(_item, slice):
                if _item.step is not None:
                    msg = "Slicing with a step is not supported."
                    raise ValueError(msg)
                return _item.start or 0
            else:
                msg = "Invalid index type"
                raise TypeError(msg)

        new_data = self._ndarray[item]
        if not isinstance(new_data, np.ndarray):
            return new_data

        if isinstance(item, tuple):
            if len(item) != 2:  # noqa: PLR2004
                msg = "Invalid number of indices"
                raise ValueError(msg)
            y_item, x_item = item

            yi = _process_item(y_item)
            xi = _process_item(x_item)
        else:
            yi = _process_item(item)
            xi = 0

        new_transform = Affine(
            self._transform.a,
            self._transform.b,
            self._transform.c + xi * self._transform.a,
            self._transform.d,
            self._transform.e,
            self._transform.f + yi * self._transform.e,
        )

        return RasterArray(new_data, new_transform, self._crs, self._no_data_value)

    # ----------------------------------------------------------------
    # NumPy array interface

    def astype(self, dtype: type[DataDtypes]) -> "RasterArray":
        """Cast the raster to a new data type."""
        return RasterArray(
            self._ndarray.astype(dtype), self._transform, self._crs, self._no_data_value
        )

    def to_numpy(self) -> RasterData:
        """Convert the raster to a NumPy array."""
        return self._ndarray.copy()

    def __array__(self, dtype: type[DataDtypes] | None = None) -> RasterData:
        return np.asarray(self._ndarray, dtype=dtype)

    def __array_ufunc__(  # noqa: C901
        self,
        ufunc: np.ufunc,
        method: NumpyUFuncMethod,
        *inputs: typing.Union[RasterData, SupportedDtypes, "RasterArray"],
        **kwargs: typing.Any,
    ) -> typing.Union[tuple["RasterArray", ...], "RasterArray"]:
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use RasterArray instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle RasterArray objects.
            handled_types = (np.ndarray, numbers.Number, RasterArray)
            if not isinstance(x, handled_types):
                return NotImplemented  # type: ignore[no-any-return]
            if isinstance(x, RasterArray):
                if x._crs != self._crs:  # noqa: SLF001
                    msg = "Coordinate reference systems do not match."
                    raise ValueError(msg)
                if not self._no_data_equal(x._no_data_value):  # noqa: SLF001
                    msg = "No data values do not match."
                    raise ValueError(msg)
                if x._transform != self._transform:  # noqa: SLF001
                    msg = "Affine transforms do not match."
                    raise ValueError(msg)

        # Propagate the no_data_value to the output array.
        no_data_mask = self.no_data_mask
        for x in inputs:
            if isinstance(x, RasterArray):
                no_data_mask |= x.no_data_mask

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x._ndarray if isinstance(x, RasterArray) else x for x in inputs)  # noqa: SLF001
        if out:
            kwargs["out"] = tuple(
                x._ndarray if isinstance(x, RasterArray) else x  # noqa: SLF001
                for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)
        result[no_data_mask] = self._no_data_value

        if type(result) is tuple:
            # multiple return values
            return tuple(
                type(self)(x, self._transform, self._crs, self._no_data_value)
                for x in result
            )
        else:
            # one return value
            return type(self)(result, self._transform, self._crs, self._no_data_value)

    # ----------------------------------------------------------------
    # Specialized array methods

    def all(self) -> bool:
        """Return True if all elements evaluate to True."""
        return self._ndarray.all()  # type: ignore[return-value]

    def any(self) -> bool:
        """Return True if any element evaluates to True."""
        return self._ndarray.any()  # type: ignore[return-value]

    # ----------------------------------------------------------------
    # Raster data

    @property
    def transform(self) -> Affine:
        """Affine transform to georeference the raster."""
        return self._transform

    @property
    def x_min(self) -> float:
        """Minimum x coordinate."""
        return self.transform.c  # type: ignore[no-any-return]

    @property
    def x_max(self) -> float:
        """Maximum x coordinate."""
        return self.x_min + self.x_resolution * self.width

    @property
    def y_min(self) -> float:
        """Minimum y coordinate."""
        return self.y_max + self.y_resolution * self.height

    @property
    def y_max(self) -> float:
        """Maximum y coordinate."""
        return self.transform.f  # type: ignore[no-any-return]

    @property
    def width(self) -> int:
        """Width of the raster."""
        return self._ndarray.shape[1]

    @property
    def height(self) -> int:
        """Height of the raster."""
        return self._ndarray.shape[0]

    @property
    def x_resolution(self) -> float:
        """Resolution in x direction."""
        return self.transform.a  # type: ignore[no-any-return]

    @property
    def y_resolution(self) -> float:
        """Resolution in y direction."""
        return self.transform.e  # type: ignore[no-any-return]

    @property
    def resolution(self) -> tuple[float, float]:
        """Resolution in x and y directions."""
        return self.x_resolution, self.y_resolution

    def x_coordinates(
        self, *, center: bool = False
    ) -> npt.NDArray[np.floating[typing.Any]]:
        """x coordinates of the raster."""
        if center:
            return np.linspace(
                self.x_min + self.x_resolution / 2,
                self.x_max - self.x_resolution / 2,
                self.width,
            )
        else:
            return np.linspace(
                self.x_min,
                self.x_max - self.x_resolution,
                self.width,
            )

    def y_coordinates(
        self, *, center: bool = False
    ) -> npt.NDArray[np.floating[typing.Any]]:
        """y coordinates of the raster."""
        if center:
            return np.linspace(
                self.y_min - self.y_resolution / 2,
                self.y_max + self.y_resolution / 2,
                self.height,
            )
        else:
            return np.linspace(
                self.y_min - self.y_resolution,
                self.y_max,
                self.height,
            )

    @property
    def bounds(self) -> tuple[float, float, float, float]:
        """Bounding box of the raster."""
        return self.x_min, self.x_max, self.y_min, self.y_max

    @property
    def crs(self) -> str | None:
        """Coordinate reference system."""
        if isinstance(self._crs, CRS):
            return self._crs.to_string()  # type: ignore[no-any-return]
        else:
            return self._crs

    def set_crs(self, new_crs: RawCRS) -> "RasterArray":
        if self._crs is not None:
            msg = (
                "Coordinate reference system is already set. Use to_crs() to reproject "
                "to a new coordinate reference system."
            )
            raise ValueError(msg)
        return RasterArray(
            self._ndarray.copy(), self._transform, new_crs, self._no_data_value
        )

    def to_crs(self, new_crs: str, resampling: str = "nearest") -> "RasterArray":
        """Reproject the raster to a new coordinate reference system."""
        if self._crs is None:
            msg = "Coordinate reference system is not set."
            raise ValueError(msg)
        return self.reproject(
            dst_crs=new_crs,
            resampling=resampling,
        )

    @property
    def no_data_value(self) -> SupportedDtypes:
        """Value representing no data."""
        if self._no_data_value is NO_DATA_UNSET:
            msg = "No data value is not set."
            raise ValueError(msg)
        return self._no_data_value

    def set_no_data_value(self, new_no_data_value: SupportedDtypes) -> "RasterArray":
        new_data = self._ndarray.copy()
        if self._no_data_value is not NO_DATA_UNSET:
            new_data[self.no_data_mask] = new_no_data_value
        return RasterArray(new_data, self._transform, self._crs, new_no_data_value)

    def _no_data_equal(self, other_no_data_value: SupportedDtypes | None) -> bool:
        if self._no_data_value is NO_DATA_UNSET:
            return other_no_data_value is NO_DATA_UNSET
        elif other_no_data_value is NO_DATA_UNSET:
            return False
        elif np.isnan(self._no_data_value):
            return np.isnan(other_no_data_value)  # type: ignore[no-any-return]
        elif np.isinf(self._no_data_value):
            other_inf = np.isinf(other_no_data_value)
            sign_match = np.sign(self._no_data_value) == np.sign(other_no_data_value)
            return other_inf and sign_match  # type: ignore[no-any-return]
        else:
            return self._no_data_value == other_no_data_value

    def unset_no_data_value(self) -> "RasterArray":
        """Unset value representing no data."""
        return RasterArray(
            self._ndarray.copy(), self._transform, self._crs, NO_DATA_UNSET
        )

    @property
    def no_data_mask(self) -> RasterMask:
        """Mask representing no data."""
        if self._no_data_value is NO_DATA_UNSET:
            return np.zeros_like(self._ndarray, dtype=bool)
        elif np.isnan(self._no_data_value):
            return np.isnan(self._ndarray)
        elif np.isinf(self._no_data_value):
            return np.isinf(self._ndarray)
        else:
            return np.equal(self._ndarray, self._no_data_value)

    def resample(self, scale: float, resampling: str = "nearest") -> "RasterArray":
        """Resample the raster."""
        dest_width = int(self.width * scale)
        dest_height = int(self.height * scale)
        destination = np.empty((dest_height, dest_width), dtype=self._ndarray.dtype)
        return self.reproject(
            destination=destination,
            dst_crs=self._crs,
            resampling=resampling,
        )

    def resample_to(
        self, target: "RasterArray", resampling: str = "nearest"
    ) -> "RasterArray":
        """Resample the raster to match the resolution of another raster."""
        destination = np.empty_like(target._ndarray, dtype=self._ndarray.dtype)  # noqa: SLF001
        return self.reproject(
            destination=destination,
            dst_transform=target.transform,
            dst_crs=target._crs,  # noqa: SLF001
            resampling=resampling,
        )

    def reproject(
        self,
        destination: RasterData | None = None,
        dst_transform: Affine | None = None,
        dst_resolution: float | tuple[float, float] | None = None,
        dst_crs: RawCRS | None = None,
        resampling: str = "nearest",
    ) -> "RasterArray":
        """Reproject the raster to match the resolution of another raster."""
        resampling = _RESAMPLING_MAP[resampling]

        dst_crs = self._crs if dst_crs is None else CRS.from_user_input(dst_crs)
        new_data, transform = reproject(
            source=self._ndarray,
            src_transform=self._transform,
            src_crs=self._crs,
            src_nodata=self._no_data_value,
            destination=destination,
            dst_transform=dst_transform,
            dst_resolution=dst_resolution,
            dst_crs=dst_crs,
            resampling=resampling,
        )
        if len(new_data.shape) == 3:  # noqa: PLR2004
            # Some operations assume and prepend a channel dimension
            new_data = new_data[0]
        return RasterArray(
            new_data,
            transform,
            dst_crs,
            self.no_data_value,
        )

    def _coerce_to_shapely(
        self, shape: Polygon | MultiPolygon | gpd.GeoDataFrame | gpd.GeoSeries
    ) -> Polygon | MultiPolygon:
        if isinstance(shape, (gpd.GeoDataFrame, gpd.GeoSeries)):
            if shape.crs != self._crs:
                msg = "Coordinate reference systems do not match."
                raise ValueError(msg)
            return shape.geometry.unary_union
        return shape

    def clip(
        self, shape: Polygon | MultiPolygon | gpd.GeoDataFrame | gpd.GeoSeries
    ) -> "RasterArray":
        """Clip the raster to a shape."""
        shape = self._coerce_to_shapely(shape)
        _, transform, window = raster_geometry_mask(
            data_transform=self.transform,
            data_width=self._ndarray.shape[1],
            data_height=self._ndarray.shape[0],
            shapes=[shape],
            crop=True,
        )

        x_start, x_end = window.col_off, window.col_off + window.width
        y_start, y_end = window.row_off, window.row_off + window.height
        new_data = self._ndarray[y_start:y_end, x_start:x_end].copy()
        return RasterArray(
            new_data, transform, self._crs, no_data_value=self.no_data_value
        )

    def mask(
        self,
        shape: Polygon | MultiPolygon | gpd.GeoDataFrame | gpd.GeoSeries,
        *,
        fill_value: SupportedDtypes | None = None,
        all_touched: bool = False,
        invert: bool = False,
    ) -> "RasterArray":
        """Mask the raster with a shape."""
        shape = self._coerce_to_shapely(shape)
        if fill_value is None and self._no_data_value is NO_DATA_UNSET:
            msg = "No fill value is set."
            raise ValueError(msg)

        if fill_value is None:
            fill_value = self.no_data_value

        shape_mask, *_ = raster_geometry_mask(
            data_transform=self.transform,
            data_width=self._ndarray.shape[1],
            data_height=self._ndarray.shape[0],
            shapes=[shape],
            all_touched=all_touched,
            invert=invert,
        )
        new_data = self._ndarray.copy()
        new_data[shape_mask] = fill_value

        return RasterArray(
            new_data, self.transform, self._crs, no_data_value=self.no_data_value
        )

    def select(
        self,
        x_coordinates: npt.NDArray[np.float64],
        y_coordinates: npt.NDArray[np.float64],
        method: str = "nearest",
    ) -> npt.NDArray[DataDtypes]:
        """Select values at specific coordinates."""
        if x_coordinates.size != y_coordinates.size:
            msg = "x and y coordinates must have the same size."
            raise ValueError(msg)

        if method == "nearest":
            x_indices = np.clip(
                np.searchsorted(self.x_coordinates(), x_coordinates),
                0,
                self.width - 1,
            )
            y_indices = np.searchsorted(
                self.y_coordinates(), y_coordinates, side="right"
            )
            # Flip y indices to match raster coordinatesok
            y_indices = np.clip(
                self.height - y_indices,
                0,
                self.height - 1,
            )

            return self._ndarray[y_indices, x_indices].copy()
        else:
            msg = "Only 'nearest' method is supported."
            raise NotImplementedError(msg)

    def __repr__(self) -> str:
        out = "RasterArray\n"
        out += "===========\n"
        out += f"dimensions    : {self.width}, {self.height} (x, y)\n"
        out += f"resolution    : {self.transform.a}, {self.transform.e} (x, y)\n"
        bounds = ", ".join(
            str(s) for s in [self.x_min, self.x_max, self.y_min, self.y_max]
        )
        out += f"extent        : {bounds} (xmin, xmax, ymin, ymax)\n"
        out += f"crs           : {self.crs}\n"
        out += f"no_data_value : {self._no_data_value}\n"
        out += f"size          : {self.nbytes / 1024 ** 2:.2f} MB\n"
        out += f"dtype         : {self._ndarray.dtype}\n"
        return out

    def to_file(self, path: FilePath, **kwargs: typing.Any) -> None:
        """Write the raster to a file."""
        from rasterra._io import write_raster

        write_raster(self, path, **kwargs)

    def to_gdf(self) -> gpd.GeoDataFrame:
        return to_gdf(self)

    @property
    def plot(self) -> Plotter:
        return Plotter(self._ndarray, self.no_data_mask, self.transform)



from typing import Any

import affine
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, Normalize
from mpl_toolkits.axes_grid1 import make_axes_locatable
from rasterio.plot import plotting_extent

from rasterra._typing import RasterData, RasterMask


class Plotter:
    def __init__(
        self,
        data: RasterData,
        data_mask: RasterMask,
        transform: affine.Affine,
    ) -> None:
        self._data = data
        self._mask = data_mask
        self._transform = transform

    def __call__(
        self,
        ax: Axes | None = None,
        cmap: str | Colormap = "viridis",
        vmin: float | None = None,
        vmax: float | None = None,
        under_color: str | None = None,
        no_data_color: str | None = None,
        norm: Normalize | None = None,
        *,
        colorbar: bool = True,
        **kwargs: Any,
    ) -> Axes:
        if (vmin is not None or vmax is not None) and norm is not None:
            msg = "Cannot pass both vmin/vmax and norm."
            raise ValueError(msg)
        if norm is None:
            norm = Normalize(vmin=vmin, vmax=vmax)

        kwargs["extent"] = plotting_extent(self._data, self._transform)

        show_plt = False
        if not ax:
            show_plt = True
            ax = plt.gca()

        _make_image_plot(
            data=self._data,
            mask=self._mask,
            norm=norm,
            cmap=cmap,
            under_color=under_color,
            no_data_color=no_data_color,
            colorbar=colorbar,
            ax=ax,
            **kwargs,
        )

        if show_plt:
            plt.show()

        return ax

    def test_normalization(
        self,
        norm: Normalize,
        nbins: int = 50,
        cmap: str | Colormap = "viridis",
        under_color: str | None = None,
        no_data_color: str | None = None,
    ) -> None:
        """Test a normalization of the data for plotting."""
        vmin = norm.vmin if norm.vmin is not None else self._data.min()
        vmax = norm.vmax if norm.vmax is not None else self._data.max()
        mask = (vmin <= self._data) & (self._data <= vmax) & ~self._mask
        data = self._data[mask].flatten()

        result = norm(data)

        fig, axes = plt.subplots(figsize=(15, 15), ncols=2, nrows=2)

        _make_hist_plot(
            data=data,
            nbins=nbins,
            title="Raw data",
            ax=axes[0, 0],
        )
        _make_image_plot(
            data=data,
            mask=self._mask,
            norm=Normalize(vmin=vmin, vmax=vmax),
            cmap=cmap,
            under_color=under_color,
            no_data_color=no_data_color,
            colorbar=False,
            ax=axes[1, 0],
        )

        _make_hist_plot(
            data=result,
            nbins=nbins,
            title="Normalized data",
            ax=axes[0, 1],
        )
        _make_image_plot(
            data=result,
            mask=self._mask,
            norm=norm,
            cmap=cmap,
            under_color=under_color,
            no_data_color=no_data_color,
            colorbar=False,
            ax=axes[1, 1],
        )

        plt.show()


def _make_hist_plot(
    data: RasterData,
    nbins: int,
    title: str,
    ax: Axes,
) -> Axes:
    """Plot a histogram of the data."""
    hist, bins = np.histogram(data, nbins)
    cdf = hist.cumsum()
    cdf = cdf / cdf.max()
    ax.bar(bins[:-1], hist, width=bins[1:] - bins[:-1], color="tab:blue")
    ax.set_ylabel("Frequency", color="tab:blue", fontsize=14)
    ax.set_title(title, fontsize=16)
    ax2 = ax.twinx()
    ax2.plot(bins[:-1], cdf, color="tab:orange")
    ax2.set_ylabel("Cumulative frequency", color="tab:orange", fontsize=14)
    return ax


def _make_image_plot(
    data: RasterData,
    mask: RasterMask,
    norm: Normalize,
    cmap: str | Colormap,
    under_color: str | None,
    no_data_color: str | None,
    ax: Axes,
    *,
    colorbar: bool,
    **kwargs: Any,
) -> Axes:
    """Plot a histogram of the data."""
    masked = np.ma.masked_array(data, mask=mask)  # type: ignore[no-untyped-call,var-annotated]
    im = ax.imshow(
        masked,
        cmap=cmap,
        norm=norm,
        **kwargs,
    )
    im.cmap.set_extremes(
        bad=no_data_color,
        under=under_color,
    )

    if colorbar:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        ax.figure.colorbar(ax.images[0], cax=cax, orientation="vertical")

    return ax



from collections.abc import Sequence
from typing import Any, TypeAlias

import rasterio
from rasterio.merge import merge
from rasterio.windows import from_bounds
from shapely import Polygon

from rasterra._array import RasterArray
from rasterra._typing import FilePath

Bounds: TypeAlias = (
    tuple[float, float, float, float] | tuple[int, int, int, int] | Polygon
)


def load_raster(
    path: FilePath,
    bounds: Bounds | None = None,
) -> RasterArray:
    """Load a raster from a file."""

    with rasterio.open(path) as f:
        if bounds is not None:
            if isinstance(bounds, Polygon):
                bounds = bounds.bounds
            window = from_bounds(*bounds, transform=f.transform)
            data = f.read(window=window, boundless=True)
            transform = f.window_transform(window)
        else:
            data = f.read()
            transform = f.transform

        if data.shape[0] == 1:
            return RasterArray(
                data[0],
                transform=transform,
                crs=f.crs,
                no_data_value=f.nodata,
            )
        else:
            msg = "Only single-band rasters are supported"
            raise NotImplementedError(msg)


def write_raster(
    raster: RasterArray,
    path: FilePath,
    **kwargs: Any,
) -> None:
    """Write a raster to a file."""
    meta = {
        "driver": "GTiff",
        "height": raster.shape[0],
        "width": raster.shape[1],
        "count": 1,
        "dtype": raster.dtype,
        "crs": raster.crs,
        "transform": raster.transform,
        "nodata": raster.no_data_value,
        **kwargs,
    }

    with rasterio.open(path, "w", **meta) as f:
        f.write(raster.to_numpy(), 1)


def load_mf_raster(paths: Sequence[FilePath]) -> RasterArray:
    """Load multiple files into a single raster."""
    with rasterio.open(paths[0]) as f:
        data = f.read()
        if data.shape[0] == 1:
            merged, transform = merge(paths)
            return RasterArray(
                merged[0],
                transform=transform,
                crs=f.crs,
                no_data_value=f.nodata,
            )
        else:
            msg = "Only single-band rasters are supported."
            raise NotImplementedError(msg)



import math
import typing

import geopandas as gpd
import numpy as np
from affine import Affine
from rasterio.features import bounds, rasterize
from rasterio.windows import Window
from shapely import box
from shapely.geometry import MultiPolygon, Polygon

from rasterra._typing import RasterData

if typing.TYPE_CHECKING:
    from rasterra import RasterArray


def _geometry_window(
    data_transform: Affine,
    data_width: int,
    data_height: int,
    shapes: list[Polygon | MultiPolygon],
    pad_x: float = 0.0,
    pad_y: float = 0.0,
) -> Window:
    all_bounds = [bounds(shape, transform=~data_transform) for shape in shapes]

    cols = [
        x
        for (left, bottom, right, top) in all_bounds
        for x in (left - pad_x, right + pad_x, right + pad_x, left - pad_x)
    ]
    rows = [
        y
        for (left, bottom, right, top) in all_bounds
        for y in (top - pad_y, top - pad_y, bottom + pad_y, bottom + pad_y)
    ]

    row_start, row_stop = int(math.floor(min(rows))), int(math.ceil(max(rows)))
    col_start, col_stop = int(math.floor(min(cols))), int(math.ceil(max(cols)))

    window = Window(
        col_off=col_start,
        row_off=row_start,
        width=max(col_stop - col_start, 0),
        height=max(row_stop - row_start, 0),
    )

    # Make sure that window overlaps raster
    raster_window = Window(0, 0, data_width, data_height)
    return window.intersection(raster_window)


def _window_transform(
    window: Window,
    transform: Affine,
) -> Affine:
    x, y = transform * (window.col_off or 0.0, window.row_off or 0.0)
    return Affine.translation(x - transform.c, y - transform.f) * transform


def raster_geometry_mask(
    data_transform: Affine,
    data_width: int,
    data_height: int,
    shapes: list[Polygon | MultiPolygon],
    *,
    all_touched: bool = False,
    invert: bool = False,
    crop: bool = False,
    pad_x: float = 0.0,
    pad_y: float = 0.0,
) -> tuple[RasterData, Affine, Window]:
    if crop and invert:
        msg = "Crop and invert cannot both be True"
        raise ValueError(msg)

    window = _geometry_window(
        data_transform,
        data_width,
        data_height,
        shapes,
        pad_x=pad_x,
        pad_y=pad_y,
    )

    if crop:
        transform = _window_transform(window, data_transform)
        out_shape = (int(window.height), int(window.width))
    else:
        window = None
        transform = data_transform
        out_shape = (data_height, data_width)

    fill_value, mask_value = (0, 1) if invert else (1, 0)

    mask = rasterize(
        shapes,
        out_shape=out_shape,
        transform=transform,
        all_touched=all_touched,
        fill=fill_value,
        default_value=mask_value,
    ).astype(bool)

    return mask, transform, window


def to_gdf(raster: "RasterArray") -> gpd.GeoDataFrame:
    xmin = raster.x_coordinates()
    ymax = raster.y_coordinates()[::-1]
    xl, yt = np.meshgrid(xmin, ymax)
    xr = xl + raster.x_resolution
    yb = yt + raster.y_resolution
    pixels = box(xl.flatten(), yb.flatten(), xr.flatten(), yt.flatten())
    gdf = gpd.GeoDataFrame(
        {"value": raster.to_numpy().flatten()}, geometry=pixels, crs=raster.crs
    )
    gdf.__from_raster__ = True
    return gdf
